--------------------------------------------------------------------------------
-- Company: UNIVERSITY OF CONNECTICUT
-- Engineer: Farrukh Hijaz
--
-- Create Date:    10/17/2014
-- Design Name:    
-- Module Name:    wb_vga640x480 - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity wb_vga640x480 is
	Port ( sys_clk : in std_logic;
        reset : in std_logic;
		adr_o : out std_logic_vector(31 downto 0);
        dat_i : in std_logic_vector(31 downto 0);
        dat_o : out std_logic_vector(31 downto 0);
        ack_i : in std_logic;
        cyc_o : out std_logic;
        stb_o : out std_logic;
        we_o  : out std_logic;
		red : out std_logic_vector(2 downto 0);
        green : out std_logic_vector(2 downto 0);
        blue : out std_logic_vector(1 downto 0);
        hsync : out std_logic;
        vsync : out std_logic;
		sw : in std_logic_vector(7 downto 0)
			);
end wb_vga640x480;

architecture Behavioral of wb_vga640x480 is
	constant VISIBLE_LINES_PER_FRAME : integer := 480;
	constant LINES_PER_FRAME : integer := 521;
	constant VSYNC_FRONT_PORCH : integer := 10;
	constant VSYNC_WIDTH : integer := 2;
	constant VSYNC_BACK_PORCH : integer := 29;
	constant TOP_VISIBLE_LINE : integer := VSYNC_FRONT_PORCH + VSYNC_WIDTH + VSYNC_BACK_PORCH;
	constant BOTTOM_VISIBLE_LINE : integer := TOP_VISIBLE_LINE + VISIBLE_LINES_PER_FRAME - 1;

	constant VISIBLE_PIXELS_PER_LINE : integer := 640;
	constant PIXELS_PER_LINE : integer := 800;
	constant HSYNC_FRONT_PORCH : integer := 16;
	constant HSYNC_WIDTH : integer := 96;
	constant HSYNC_BACK_PORCH : integer := 48;
	constant LEFT_VISIBLE_PIXEL : integer := HSYNC_FRONT_PORCH + HSYNC_WIDTH + HSYNC_BACK_PORCH;
	constant RIGHT_VISIBLE_PIXEL : integer := LEFT_VISIBLE_PIXEL + VISIBLE_PIXELS_PER_LINE - 1;

	constant CHARS_PER_LINE : integer := 80;
	constant LINES_PER_PAGE : integer := 40;
	constant CHARS_PER_PAGE : integer := CHARS_PER_LINE * LINES_PER_PAGE;
	constant CHAR_WIDTH : integer := 8;
	constant CHAR_HEIGHT : integer := 12;
	constant PIXELS_PER_WORD : integer := 8;
	constant BITS_PER_PIXEL : integer := 4;

	signal pixel_clk : std_logic;
	signal line_count : integer range 0 to LINES_PER_FRAME;
	signal pixel_count : integer range 0 to PIXELS_PER_LINE;
	signal hsync_internal : std_logic;
	signal display_valid, hdisplay_valid, vdisplay_valid : std_logic;

	signal pixel : std_logic_vector(BITS_PER_PIXEL-1 downto 0);
	signal pixnum : integer range 0 to PIXELS_PER_WORD-1;
	signal offset : integer range 0 to VISIBLE_PIXELS_PER_LINE;
	signal line_offset : integer range 0 to VISIBLE_LINES_PER_FRAME;

	signal txtcolor, bgcolor : std_logic_vector(BITS_PER_PIXEL-1 downto 0);
	signal buffer_base : std_logic_vector(31 downto 0) := X"40000000";
	
begin

	bgcolor <= sw(7 downto 4);

	clk_divider : entity work.clock_divider
		generic map ( divisor => 2 )
		port map ( clk_in => sys_clk, reset => reset, clk_out => pixel_clk );


	line_offset <= line_count - TOP_VISIBLE_LINE; -- Points to the line number in the visible region
	offset <= pixel_count-LEFT_VISIBLE_PIXEL; -- Points to the pixel number in the visible region
	pixnum <= offset mod PIXELS_PER_WORD; -- Points to the pixel number within a 32-bits word
	-- pixel should be read from the stored character if the offset and line_offset are within the range of the 1st character (i.e. first 8 and 12 respectively), bgcolor otherwise.
	
	vdisplay_valid <= '1' when (line_count >= TOP_VISIBLE_LINE and line_count <= BOTTOM_VISIBLE_LINE) else '0';
	hdisplay_valid <= '1' when (pixel_count >= LEFT_VISIBLE_PIXEL and pixel_count <= RIGHT_VISIBLE_PIXEL) else '0';
	display_valid <= hdisplay_valid and vdisplay_valid;

	red <= pixel(3 downto 2) & "0" when display_valid = '1' else "000";
	green <= pixel(1) & "00" when display_valid = '1' else "000";
	blue <= pixel(0) & "0" when display_valid = '1' else "00";

	hsync <= '0' when pixel_count>=HSYNC_FRONT_PORCH and pixel_count < HSYNC_FRONT_PORCH+HSYNC_WIDTH else '1';
	vsync <= '0' when (line_count >= VSYNC_FRONT_PORCH and line_count < VSYNC_FRONT_PORCH+VSYNC_WIDTH) else '1';
	
	pixel_count_process: process( pixel_clk, reset )
	begin
		if ( reset = '1' ) then
			pixel_count <= 0;
		elsif ( pixel_clk'event and pixel_clk='1' ) then
			
			if ( pixel_count = PIXELS_PER_LINE-1 ) then
				pixel_count <= 0;
			else
				pixel_count <= pixel_count + 1;
			end if;

		end if;
	end process;
	
	line_count_process: process( pixel_clk, reset )
	begin
		if ( reset = '1' ) then
			line_count <= 0;
		elsif ( pixel_clk'event and pixel_clk='1' and pixel_count = 0) then
			if ( line_count = LINES_PER_FRAME-1 ) then
				line_count <= 0;
			else
				line_count <= line_count + 1;
			end if;
		end if;
	end process;



end Behavioral;
